{******************************************************************************}
{*                                                                            *}
{* Component: NeuralNetwork Back Propagation                                  *}
{*                                                                            *}
{* Author: Miktyuk Sergey, Frolov Valentin aka 'BonifacE'                     *}
{* tel: +7-0912-25-17-50, +7-0912-25-17-15                                    *}
{* Email: mail@silentium.ru                                                   *}
{* Copyright (c) Silentium 2004                                               *}
{******************************************************************************}

unit NeuralNetworkBP;

interface

uses IniFiles, SysUtils, Classes;

type
 Neuron = record                            // 
           Weights: array of double;        //  
           DeltaWeights: array of double;   //  
           Value: double;                   // 
           Sigma: double;                   // 
          end;

 Layer = record                             // 
          CountNeurons: integer;            //   
          Neurons: array of Neuron;         // 
          HiddenNeuronWeight: array of double;
          HiddenNeuronDeltaWeight: array of double;
          HiddenNeuronSigma: double;
         end;

// TNeuralNetwork = object
 TNeuralNetworkBP = class (TComponent)
 private
  fNN: array of Layer;               // 
                                     //  () 
                                     //    
                                     //(  )

  fTM: array of array of double;     // 
  fTNM: array of array of double;    //  
  fNM: array of array of double;     //  
  fIM: array of array of double;     //   
  fOM: array of array of double;     //   
  fVTM: array of array of double;
  fEM: array of double;              //  
  fE: double;                        //    
  fCH: integer;                      //  
  fCHL: integer;                     //  
  fCE: integer;                      //  
  fAlpha: double;                    //  (0..unl)
  fTS: double;                       //  (0..1)
  fMiu: double;                      //  (0..1)
  fTL: double;                       // 
  fCCV: integer;                     // . 
  fMEC: integer;
  fMSC: integer;
  fINNC, fOUTNC, fEpC, fShC: integer;
  function fGetIM(i,j:integer):double;
  function fGetOM(i,j:integer):double;
  function fGetTM(i,j:integer):double;
  function fGetVTM(i,j:integer):double;

  procedure fSetIM(i,j:integer;Val:double);
  procedure fSetOM(i,j:integer;Val:double);
  procedure fSetTM(i,j:integer;Val:double);
  procedure fSetVTM(i,j:integer;Val:double);

  procedure GetRes(nv: integer);

  procedure Denormalization;
  procedure CalculateAxons;
  procedure CalculateError;
  procedure CalculateSigma;
  procedure WeightsCorrection;
  procedure Shock;
  procedure HebbTeach;

 public
  procedure Normalization;
  //   
  property InMatrix[i,j:integer]: double read fGetIM write fSetIM;
  //   
  property OutMatrix[i,j:integer]: double read fGetOM write fSetOM;
  // 
  property TeachMatrix[i,j:integer]: double read fGetTM write fSetTM;
  //  
  property CountHiddenLayers: integer read fCHL write fCHL;
  //    
  property VTMatrix[i,j:integer]: double read fGetVTM write fSetVTM;
  property Epsilon: double read fE write fE;
  property OutNC: integer read fOUTNC write fOUTNC;
  property InNC: integer read fINNC write fINNC;
  procedure Configurate(HLNC: array of integer);
  procedure SaveNetParam(FN: TFileName);
  procedure LoadNetParam(FN: TFileName);
  function Learn: boolean;
  procedure Calculate;
  procedure SetCalcMatrix(CountCalcVector:integer);
  procedure SetTeachMatrix(CountExamples:integer);
  constructor Create(AOwner: TComponent); override;


 published
  //  (0..unl)
  property Alpha: double read fAlpha write fAlpha;
  //  (0..1)
  property TeachSpeed: double read fTS write fTS;
    // 
  property Tolerance: double read fTL write fTL;
  //  (0..1)
  property Miu: double read fMiu write fMiu;
  //.  
  property MaxEpochCount: integer read fMEC write fMEC;
  //.   
  property MaxShockCount: integer read fMSC write fMSC;
 end;

 procedure Register;

implementation

{$R *.RES}

procedure Register;
begin
  RegisterComponents('Additional', [TNeuralNetworkBP]);
end;

function TNeuralNetworkBP.fGetIM(i,j:integer):double;
begin
  result := fIM[i,j];
end;

procedure TNeuralNetworkBP.fSetIM;
begin
  fIM[i,j]:=Val;
end;

function TNeuralNetworkBP.fGetOM(i,j:integer):double;
begin
  result := fOM[i,j];
end;

procedure TNeuralNetworkBP.fSetOM;
begin
  fOM[i,j]:=Val;
end;

function TNeuralNetworkBP.fGetTM(i,j:integer):double;
begin
  result := fTM[i,j];
end;

procedure TNeuralNetworkBP.fSetTM;
begin
  fTM[i,j]:=Val;
end;

function TNeuralNetworkBP.fGetVTM(i,j:integer):double;
begin
  result := fVTM[i,j];
end;

procedure TNeuralNetworkBP.fSetVTM;
begin
  fVTM[i,j]:=Val;
end;

constructor TNeuralNetworkBP.Create(AOwner: TComponent);
begin
 inherited;
 Miu:=0.5;
 Alpha:=1;
 TeachSpeed:=0.95;
 Tolerance:=0.05;
 MaxEpochCount:=10000;
 MaxShockCount:=3;
 fEpC:=0;
 fShC:=0;
end;

procedure TNeuralNetworkBP.Configurate(HLNC: array of integer);
var
 i,j,k:integer;
begin
//--  -------------------------------------------------------//
 SetLength(fNN,fCHL+2);
 for i:=0 to fCHL+1 do
  begin
   SetLength(fNN[i].Neurons,HLNC[i]);
   fNN[i].CountNeurons:=HLNC[i];
  end;
 for i:=1 to fCHL+1 do
  for j:=0 to fNN[i].CountNeurons-1 do
   begin
    SetLength(fNN[i].Neurons[j].Weights,fNN[i-1].CountNeurons);
    SetLength(fNN[i].Neurons[j].DeltaWeights,fNN[i-1].CountNeurons);
   end;
 for i:=fCHL downto 1 do
  begin
   SetLength(fNN[i].HiddenNeuronWeight,fNN[i+1].CountNeurons);
   SetLength(fNN[i].HiddenNeuronDeltaWeight,fNN[i+1].CountNeurons);
  end;

//----------------------------------------------------------------------------//

//---   ---------------------------------//
 for i:=fCHL+1 downto 1 do
  begin
   if (i<fCHL+1)
    then
     for j:=0 to fNN[i+1].CountNeurons-1 do
      fNN[i].HiddenNeuronWeight[j]:= 0.2;//(Random(99)+1)/100;
   for j:=0 to fNN[i].CountNeurons-1 do
    begin
     fNN[i].Neurons[j].Value:=0.0;
     fNN[i].Neurons[j].Sigma:=0.0;
     for k:=0 to fNN[i-1].CountNeurons-1 do
      fNN[i].Neurons[j].Weights[k]:= (Random(99)+1)/100;
    end;
  end;
//----------------------------------------------------------------------------//
end;

procedure TNeuralNetworkBP.SetTeachMatrix(CountExamples: integer);
var
 i:integer;
begin
 fCE:=CountExamples;
 SetLength(fEM,fCE);
 SetLength(fTM,fCE);
 SetLength(fVTM,fCE);
 SetLength(fTNM,fCE);
 for i:=0 to fCE-1 do
  begin
   SetLength(fTM[i],fNN[0].CountNeurons+fNN[fCHL+1].CountNeurons);
   SetLength(fTNM[i],fNN[0].CountNeurons+fNN[fCHL+1].CountNeurons);
   SetLength(fVTM[i],fNN[0].CountNeurons+fNN[fCHL+1].CountNeurons+1);
  end;
 SetLength(fNM,2);
 for i:=0 to 1 do
  SetLength(fNM[i],fNN[0].CountNeurons+fNN[fCHL+1].CountNeurons);

end;

procedure TNeuralNetworkBP.SetCalcMatrix(CountCalcVector: integer);
var
 i:integer;
begin
 fCCV:=CountCalcVector;
 SetLength(fIM,fCCV);
 for i:=0 to fCCV-1 do
  SetLength(fIM[i],fNN[0].CountNeurons);
 SetLength(fOM,fCCV);
 for i:=0 to fCCV-1 do
  SetLength(fOM[i],fNN[fCHL+1].CountNeurons);
end;

procedure TNeuralNetworkBP.Shock;
var
 i,j,k: integer;
begin
 // (     )
 for i:=1 to fCHL+1 do
  for j:=0 to fNN[i].CountNeurons-1 do
    for k:=0 to fNN[i-1].CountNeurons-1 do
     fNN[i].Neurons[j].Weights[k]:=fNN[i].Neurons[j].Weights[k]+Random*0.5-0.25;
end;

procedure TNeuralNetworkBP.Normalization;
var
 i,j: integer;
 min,max: double;
begin
 //   
 for j:= 0 to fNN[0].CountNeurons+fNN[fCHL+1].CountNeurons-1 do
  begin
   min:=fTM[0,j];
   max:=fTM[0,j];
   fNM[0,j]:=min;
   fNM[1,j]:=max;
   for i:= 0 to fCE-1 do
    begin
     if min>fTM[i,j]
      then
       begin
        min:=fTM[i,j];
        fNM[0,j]:=min;
        end;
     if max<fTM[i,j]
      then
       begin
        max:=fTM[i,j];
        fNM[1,j]:=max;
       end;
    end;
  end;

 //  
 for j:= 0 to fNN[0].CountNeurons+fNN[fCHL+1].CountNeurons-1 do
  for i:= 0 to fCE-1 do
   fTNM[i,j]:=(fTM[i,j]-fNM[0,j])/(fNM[1,j]-fNM[0,j]);
end;

procedure TNeuralNetworkBP.Denormalization;
var
 i,j:integer;
begin
 //    
 for i:=0 to fCCV-1 do
  for j:=0 to fNN[fCHL+1].CountNeurons-1 do
   fOM[i,j]:=fOM[i,j]*(fNM[1,fNN[0].CountNeurons+j]-fNM[0,fNN[0].CountNeurons+j])+fNM[0,fNN[0].CountNeurons+j];
end;

procedure TNeuralNetworkBP.CalculateAxons;
var
 i,j,k: integer;
 sum: double;
begin
 for i:=1 to fCHL+1 do
  for j:=0 to fNN[i].CountNeurons-1 do
   begin
    sum:=0;
    //    (  )
    for k:=0 to fNN[i-1].CountNeurons-1 do
     sum:=sum+fNN[i-1].Neurons[k].Value*fNN[i].Neurons[j].Weights[k];
    //     ()
    if (i>1) then
     sum:=sum+fNN[i-1].HiddenNeuronWeight[j];
    fNN[i].Neurons[j].Value:=1/(1+Exp(-1*fAlpha*sum));
   end;
end;

procedure TNeuralNetworkBP.CalculateError;
var
 j:integer;
 temp: double;
begin
 temp:=0;
 //  
 for j:=0 to fNN[1+fCHL].CountNeurons-1 do
  temp:=temp+Sqr(fNN[1+fCHL].Neurons[j].Value-fTNM[fCH,fNN[0].CountNeurons+j]);
 fEM[fCH]:=temp/2;
end;

procedure TNeuralNetworkBP.CalculateSigma;
var
 i,j,k: integer;
 temp: double;
begin
 temp:=0;
 //    
 for j:=0 to fNN[1+fCHL].CountNeurons-1 do
  begin
   fNN[1+fCHL].Neurons[j].sigma:=fNN[1+fCHL].Neurons[j].Value-fTNM[fCH,fNN[0].CountNeurons+j];
   temp:= temp+fNN[1+fCHL].Neurons[j].sigma;
  end;
//    
 for i:=fCHL downto 1 do
  for j:=0 to fNN[i].CountNeurons-1 do
   begin
    temp:=0;
    //      
    for k:= 0 to fNN[i+1].CountNeurons-1 do
     temp:=temp+fNN[i+1].Neurons[k].sigma*fNN[i+1].Neurons[k].Weights[j];
    //     
    fNN[i].Neurons[j].Sigma:=temp*fAlpha*fNN[i].Neurons[j].Value*(1-fNN[i].Neurons[j].Value);
   end;
end;

procedure TNeuralNetworkBP.WeightsCorrection;
var
 i,j,k: integer;
begin
 for i:=fCHL+1 downto 1 do
  for j:=0 to fNN[i].CountNeurons-1 do
   for k:= 0 to fNN[i-1].CountNeurons-1 do
    fNN[i].Neurons[j].DeltaWeights[k]:=fNN[i].Neurons[j].DeltaWeights[k]*fMiu+
     (-1*fTS*fNN[i].Neurons[j].sigma*fNN[i-1].Neurons[k].Value);

 //  
 for i:=fCHL+1 downto 1 do
  for j:=0 to fNN[i].CountNeurons-1 do
   for k:=0 to fNN[i-1].CountNeurons-1 do
    fNN[i].Neurons[j].Weights[k]:=fNN[i].Neurons[j].Weights[k]+fNN[i].Neurons[j].DeltaWeights[k];
end;

function TNeuralNetworkBP.Learn: boolean;
var
 EpsilonTemp: double;
 i,j,k,q:integer;
begin
 Result:=true;
 if (fEpC=fMEC) and (fShC<fMSC)
  then
   begin
    Shock;
    fShC:=fShC+1;
    fEpC:=0
   end;
  if (fEpC=fMEC) and (fShC=fMSC)
   then
    begin
     Result:=false;
    end;
  for i:=0 to fCE-1 do
   begin
    fCH:=i;
    for j:=0 to fNN[0].CountNeurons-1 do
     fNN[0].Neurons[j].Value:=fTNM[fCH,j];
    CalculateAxons;
    CalculateSigma;
    WeightsCorrection;
    CalculateError;
    for j:=0 to fNN[0].CountNeurons-1 do
     fVTM[i,j]:=fNN[0].Neurons[j].Value;
    for j:=0 to fNN[fCHL+1].CountNeurons-1 do
     fVTM[i,j+fNN[0].CountNeurons]:=fNN[fCHL+1].Neurons[j].Value;
    fVTM[i,fNN[0].CountNeurons+fNN[fCHL+1].CountNeurons]:=fEM[i];
   end;
  EpsilonTemp:=fEM[0];
  for i:=1 to fCE-1 do
   EpsilonTemp:=EpsilonTemp+fEM[i];
  fE:=EpsilonTemp/fCE;
  fEpC:=fEpC+1;
  for j:= 0 to fNN[0].CountNeurons+fNN[fCHL+1].CountNeurons-1 do
   for i:= 0 to fCE-1 do
    fVTM[i,j]:=fVTM[i,j]*(fNM[1,j]-fNM[0,j])+fNM[0,j];
end;

procedure TNeuralNetworkBP.Calculate;
var
 i,j: integer;
begin
 for i:=0 to fCCV-1 do
  begin
   for j:=0 to fNN[0].CountNeurons-1 do
    fNN[0].Neurons[j].Value:=(fIM[i,j]-fNM[0,j])/(fNM[1,j]-fNM[0,j]);
   CalculateAxons;
   GetRes(i);
  end;
 Denormalization;
end;

procedure TNeuralNetworkBP.GetRes(nv: integer);
var
 j:integer;
begin
 for j:=0 to fNN[fCHL+1].CountNeurons-1 do
  fOM[nv,j]:=fNN[fCHL+1].Neurons[j].Value
end;

procedure TNeuralNetworkBP.SaveNetParam(FN: TFileName);
var
 NeuroCF: TIniFile;
 i,j,k:integer;
begin
 NeuroCF:=TIniFile.Create(FN);

 NeuroCF.WriteFloat('GeneralCfg','Alpha',fAlpha);
 NeuroCF.WriteFloat('GeneralCfg','Miu',fMiu);
 NeuroCF.WriteFloat('GeneralCfg','TechSpeed',fTS);
 NeuroCF.WriteFloat('GeneralCfg','Tolerance',fTL);
 NeuroCF.WriteInteger('GeneralCfg','MaxEpochCount',fMEC);
 NeuroCF.WriteInteger('GeneralCfg','MaxShockCount',fMSC);
 NeuroCF.WriteInteger('LayersCfg','CountHiddenLayers',fCHL);
 for i:=0 to fCHL+1 do
  NeuroCF.WriteInteger('LayersCfg','Layer '+inttostr(i)+' CountNeurons',fNN[i].CountNeurons);
 for i:=fCHL+1 downto 1 do
  for j:=0 to fNN[i].CountNeurons-1 do
   for k:=0 to fNN[i-1].CountNeurons-1 do
    NeuroCF.WriteFloat('Inside parametrs','Layer '+inttostr(i)+' Neuron '+inttostr(j)+' Weight '+inttostr(k)+' ',fNN[i].Neurons[j].Weights[k]);
  for i:=0 to fNN[0].CountNeurons+fNN[fCHL+1].CountNeurons-1 do
  NeuroCF.WriteFloat('Normalization','min '+floattostr(i)+' ',fNM[0,i]);
 for i:=0 to fNN[0].CountNeurons+fNN[fCHL+1].CountNeurons-1 do
  NeuroCF.WriteFloat('Normalization','max '+floattostr(i)+' ',fNM[1,i]);
 NeuroCF.Free;
end;

procedure TNeuralNetworkBP.LoadNetParam(FN: TFileName);
var
 NeuroCF: TIniFile;
 i,j,k:integer;
 CfgM: array of integer;
begin
 NeuroCF:=TIniFile.Create(FN);
 fCHL:=NeuroCF.ReadInteger('LayersCfg','CountHiddenLayers',0);
 SetLength(CfgM,fCHL+2);
 for i:=0 to fCHL+1 do
  CfgM[i]:=NeuroCF.ReadInteger('LayersCfg','Layer '+inttostr(i)+' CountNeurons',1);
 Configurate(CfgM);
 for i:=fCHL+1 downto 1 do
  for j:=0 to fNN[i].CountNeurons-1 do
   for k:=0 to fNN[i-1].CountNeurons-1 do
    fNN[i].Neurons[j].Weights[k]:=NeuroCF.ReadFloat('Inside parametrs','Layer '+inttostr(i)+' Neuron '+inttostr(j)+' Weight '+inttostr(k)+' ',0);
 fAlpha:=NeuroCF.ReadFloat('GeneralCfg','Alpha',1);
 fMiu:=NeuroCF.ReadFloat('GeneralCfg','Miu',0.5);
 fTS:=NeuroCF.ReadFloat('GeneralCfg','TechSpeed',0.05);
 fTL:=NeuroCF.ReadFloat('GeneralCfg','Tolerance',0.05);
 fMEC:=NeuroCF.ReadInteger('GeneralCfg','MaxEpochCount',100000);
 fMSC:=NeuroCF.ReadInteger('GeneralCfg','MaxShockCount',2);
 SetLength(fNM,2);
 for i:=0 to 1 do
  SetLength(fNM[i],fNN[0].CountNeurons+fNN[fCHL+1].CountNeurons);
 for i:=0 to fNN[0].CountNeurons+fNN[fCHL+1].CountNeurons-1 do
  fNM[0,i]:=NeuroCF.ReadFloat('Normalization','min '+floattostr(i)+' ',0);
 for i:=0 to fNN[0].CountNeurons+fNN[fCHL+1].CountNeurons-1 do
  fNM[1,i]:=NeuroCF.ReadFloat('Normalization','max '+floattostr(i)+' ',0);
 fOUTNC:=fNN[fCHL+1].CountNeurons;
 fINNC:=fNN[0].CountNeurons;
 NeuroCF.Free;
end;

procedure TNeuralNetworkBP.HebbTeach;
var
 i,j,k: integer;
begin
 for i:=1 to fCHL+1 do
  for j:=0 to fNN[i].CountNeurons-1 do
   for k:=0 to fNN[i-1].CountNeurons-1 do
    fNN[i].Neurons[j].Weights[k]:=fNN[i].Neurons[j].Weights[k]+fTS*fNN[i].Neurons[j].Value*fNN[i-1].Neurons[k].Value;
end;

end.

